# -*- coding: utf-8 -*-

import mock

from django import template
from django.test import TestCase
from django.test.utils import override_settings

from .models import Address


class AddressTests(TestCase):
    fake_default_center = (1, 2)

    def test_empty_dont_save_on_db(self):
        # If we pass an empty address we don't save nothing in the
        # database
        html = "{%% load easy_maps_tags %%}{%% easy_map '%(v)s' 500 500 10 %%}"

        n_addresses_before = len(Address.objects.all())

        t = template.Template(html % {"v": ""})
        t.render(template.Context({}))

        n_addresses_after = len(Address.objects.all())

        self.assertEqual(n_addresses_after, n_addresses_before)

    @override_settings(EASY_MAPS_CENTER=fake_default_center)
    def test_empty_address_use_defaults(self):
        # When an empty address is passed uses the EASY_MAPS_CENTER
        # setting
        html = "{%% load easy_maps_tags %%}{%% easy_map '%(v)s' 500 500 10 %%}"

        address = [None]  # nonlocal

        # below we patch the render_to_string in order to retrieve the
        # map context variable and check its coordinate
        def get_address_instance(*args, **kwargs):
            template_name, context = args
            address[0] = context['map']
            return ''

        t = template.Template(html % {"v": ""})
        with mock.patch(
                'classytags.helpers.render_to_string', get_address_instance):
            t.render(template.Context({}))

        self.assertEqual(
            address[0].latitude, AddressTests.fake_default_center[0])
        self.assertEqual(
            address[0].longitude, AddressTests.fake_default_center[1])

    @override_settings(EASY_MAPS_CENTER=fake_default_center)
    def test_normal_address(self):
        # If we pass an address don't use the defaults
        html = "{%% load easy_maps_tags %%}{%% easy_map '%(v)s' 500 500 10 %%}"

        n_addresses_before = len(Address.objects.all())

        a = "Ekaterinburg, Mira 33"

        t = template.Template(html % {"v": a})
        t.render(template.Context({}))

        address = Address.objects.get(address=a)

        self.assertNotEqual(
            address.latitude, AddressTests.fake_default_center[0])
        self.assertNotEqual(
            address.longitude, AddressTests.fake_default_center[1])

        n_addresses_after = len(Address.objects.all())

        self.assertEqual(n_addresses_after, n_addresses_before + 1)

    @override_settings(EASY_MAPS_CENTER=fake_default_center)
    def test_use_address_instance(self):
        # It's possible to pass directly to the easy_map tag an
        # Address instance.  This test checks also that the database
        # is not hit.
        html = "{%% load easy_maps_tags %%}{%% easy_map %(v)s 500 500 10 %%}"

        # create a fake address
        a = Address.objects.create(address='fake')

        n_addresses_before = len(Address.objects.all())

        t = template.Template(html % {'v': 'address'})
        ctx = template.Context({'v': a})

        self.assertNumQueries(0, lambda: t.render(ctx))

        n_addresses_after = len(Address.objects.all())

        # no address is created in the process
        self.assertEqual(n_addresses_after, n_addresses_before)



# -*- coding: utf-8 -*-

import warnings

from django.conf import settings  # pylint: disable=W0611

from appconf import AppConf


class EasyMapsSettings(AppConf):
    CENTER = (-41.3, 32)
    GEOCODE = 'easy_maps.geocode.google_v3'
    ZOOM = 16  # See https://developers.google.com/maps/documentation/javascript/tutorial#MapOptions for more information.
    LANGUAGE = 'en'  # See https://developers.google.com/maps/faq#languagesupport for supported languages.
    GOOGLE_MAPS_API_KEY = None
    GOOGLE_KEY = None

    CACHE_LIFETIME = 600  # 10 minutes in seconds

    class Meta:
        prefix = 'easy_maps'
        holder = 'easy_maps.conf.settings'

if settings.EASY_MAPS_GOOGLE_MAPS_API_KEY is not None:
    warnings.warn("EASY_MAPS_GOOGLE_MAPS_API_KEY is deprecated, use EASY_MAPS_GOOGLE_KEY", DeprecationWarning)



# -*- coding: utf-8 -*-

from django.core.exceptions import ImproperlyConfigured


def importpath(path, error_text=None):
    """
    Import value by specified ``path``.
    Value can represent module, class, object, attribute or method.
    If ``error_text`` is not None and import will
    raise ImproperlyConfigured with user friendly text.

    """
    result = None
    attrs = []
    parts = path.split('.')
    exception = None
    while parts:
        try:
            result = __import__('.'.join(parts), {}, {}, [''])
        except ImportError as e:
            if exception is None:
                exception = e
            attrs = parts[-1:] + attrs
            parts = parts[:-1]
        else:
            break
    for attr in attrs:
        try:
            result = getattr(result, attr)
        except (AttributeError, ValueError) as e:
            if error_text is not None:
                raise ImproperlyConfigured('Error: %s can import "%s"' % (
                    error_text, path))
            else:
                raise exception
    return result



# -*- coding: utf-8 -*-

import sys
import logging
import collections
from datetime import timedelta

from django.db import models
from django.utils import timezone
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from . import geocode
from .conf import settings
from .utils import importpath

logger = logging.getLogger(__name__)


class AddressManager(models.Manager):

    def for_address(self, address):
        obj, created = self.get_or_create(address=address)

        if timezone.now() > obj.expiry() and not created:
            obj.save()  # trigger fetching once again

        return obj


@python_2_unicode_compatible
class Address(models.Model):

    address = models.CharField(_('address'), max_length=255, unique=True)

    # for internal use...

    computed_address = models.CharField(
        _('computed address'), max_length=255, null=True, blank=True)
    latitude = models.FloatField(_('latitude'), null=True, blank=True)
    longitude = models.FloatField(_('longitude'), null=True, blank=True)

    exception = models.TextField(_('has exception?'), blank=True)

    timestamp = models.DateTimeField(auto_now=True)

    objects = AddressManager()

    class Meta:
        verbose_name = _("EasyMaps Address")
        verbose_name_plural = _("Address Geocoding Cache")

    def __str__(self):
        return self.address

    def expiry(self, *args, **kwargs):
        return self.timestamp + timedelta(seconds=settings.EASY_MAPS_CACHE_LIFETIME)

    def fetch(self):
        if not self.address:
            return None

        func = getattr(settings, 'EASY_MAPS_GEOCODE', None)
        if func is not None:
            if not isinstance(func, collections.Callable):
                func = importpath(func)

        try:
            self.computed_address, (self.latitude, self.longitude,) = func(self.address)
        except geocode.Error as e:
            self.computed_address = self.latitude = self.longitude = None  # shit happens

            try:
                logger.error(e)
            except Exception:
                logger.error(
                    "Geocoding error for address '%s'", address)

            self.exception = "{0.__name__}: {1}".format(
                sys.exc_info()[0], sys.exc_info()[1])

    def has_exception(self):
        return bool(self.exception)
    has_exception.short_description = _("has exception?")
    has_exception.boolean = True

    def save(self, *args, **kwargs):
        if bool(self.computed_address) is False:  # just in case
            self.longitude = self.latitude = None
            
        if (self.longitude is None) or (self.latitude is None):
            self.fetch()

        super(Address, self).save(*args, **kwargs)



# Generated by Django 2.1.7 on 2019-03-29 05:41

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('easy_maps', '0001_initial'),
    ]

    operations = [
        migrations.RemoveField(
            model_name='address',
            name='geocode_error',
        ),
        migrations.AddField(
            model_name='address',
            name='exception',
            field=models.TextField(blank=True, verbose_name='has exception?'),
        ),
    ]



# -*- coding: utf-8 -*-

from django import forms
from django.contrib import admin
from django.utils.translation import ugettext_lazy as _

from .models import Address
from .widgets import AddressWithMapWidget


class HasExceptionFilter(admin.SimpleListFilter):
    title = _('exception')
    parameter_name = 'has_exception'
    
    def lookups(self, request, model_admin):
        return (
            (1, _('Yes')),
            (0, _('No')),
        )

    def queryset(self, request, queryset):
        if self.value() is not None:
            ids = Address.objects.values_list('pk', flat=True)
            
            if self.value() == '1':
                return queryset.filter(pk__in=ids)
            
            elif self.value() == '0':
                return queryset.exclude(pk__in=ids)

        return queryset


class AddressAdmin(admin.ModelAdmin):
    list_display = [
        'address', 'computed_address', 'latitude', 'longitude', 'has_exception']
    list_filter = [HasExceptionFilter]
    search_fields = ['address']

    class form(forms.ModelForm):
        class Meta:
            widgets = {
                'address': AddressWithMapWidget({'class': 'vTextField'})
            }



# Generated by Django 2.1.7 on 2019-03-29 07:13

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('easy_maps', '0002_auto_20190329_0541'),
    ]

    operations = [
        migrations.AddField(
            model_name='address',
            name='timestamp',
            field=models.DateTimeField(auto_now=True),
        ),
    ]



# -*- coding: utf-8 -*-

from django import template
from django.forms import TextInput


class AddressWithMapWidget(TextInput):
    width = 700
    height = 200
    zoom = 16

    tpl = "{{% load easy_maps_tags %}}{{% easy_map address {0.width} {0.height} {0.zoom} %}}"

    def render(self, name, value, attrs=None, renderer=None):
        output = super(AddressWithMapWidget, self).render(name, value, attrs, renderer)

        t = template.Template(self.tpl.format(self))
        context = template.Context({
            'address': value,
        })
        return output + t.render(context)



# -*- coding: utf-8 -*-

from django import template
from django.core.exceptions import ImproperlyConfigured

from classytags.core import Options
from classytags.helpers import InclusionTag
from classytags.arguments import Argument, IntegerArgument

from ..conf import settings
from ..models import Address

register = template.Library()


def parse_address(address=None):
    if isinstance(address, Address):
        return address

    if not address:
        return Address(
            latitude=settings.EASY_MAPS_CENTER[0],
            longitude=settings.EASY_MAPS_CENTER[1],
        )

    return Address.objects.for_address(address)


class EasyMapTag(InclusionTag):
    """
    The syntax:

    {% easy_map <address> [<width> <height>] [<zoom>] [using <template_name>] %}

    The "address" parameter can be an ``easy_maps.Address`` instance
    or a string describing it.  If an address is not found a new entry
    is created in the database.

    """
    name = 'easy_map'
    template = 'easy_maps/map.html'
    options = Options(
        Argument('address', resolve=True, required=True),
        Argument('width', required=False, default=None),
        Argument('height', required=False, default=None),
        IntegerArgument('zoom', required=False, default=None),
        'using',
        Argument('template_name', default=None, required=False),
    )

    def render_tag(self, context, **kwargs):
        params = dict((k, v) for k, v in kwargs.items() if v and k not in ['template_name'])
        if 'address' in params and (len(params) == 2 or len(params) > 4):
            raise template.TemplateSyntaxError(
                "easy_map tag has the following syntax: "
                "{% easy_map <address> [<width> <height>] [zoom] [using <template_name>] %}"
            )

        if settings.EASY_MAPS_GOOGLE_KEY is None and settings.EASY_MAPS_GOOGLE_MAPS_API_KEY is None:
            raise ImproperlyConfigured(
                "easy_map tag requires EASY_MAPS_GOOGLE_KEY to be set in global settings "
                "because of the restrictions introduced in Google Maps API v3 by Google, Inc."
            )
        return super(EasyMapTag, self).render_tag(context, **kwargs)

    def get_template(self, context, **kwargs):
        return kwargs.get('template_name', None) or self.template

    def get_context(self, context, **kwargs):
        kwargs.update({'map': parse_address(kwargs.pop('address'))})
        if not kwargs.get('zoom', None):
            kwargs['zoom'] = settings.EASY_MAPS_ZOOM  # default value
        kwargs['language'] = settings.EASY_MAPS_LANGUAGE
        kwargs['api_key'] = settings.EASY_MAPS_GOOGLE_KEY or settings.EASY_MAPS_GOOGLE_MAPS_API_KEY
        return kwargs

register.tag(EasyMapTag)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='Address',
            fields=[
                ('id', models.AutoField(verbose_name='ID', serialize=False, auto_created=True, primary_key=True)),
                ('address', models.CharField(unique=True, max_length=255, verbose_name='address')),
                ('computed_address', models.CharField(max_length=255, null=True, verbose_name='computed address', blank=True)),
                ('latitude', models.FloatField(null=True, verbose_name='latitude', blank=True)),
                ('longitude', models.FloatField(null=True, verbose_name='longitude', blank=True)),
                ('geocode_error', models.BooleanField(default=False, verbose_name='geocode error')),
            ],
            options={
                'verbose_name': 'EasyMaps Address',
                'verbose_name_plural': 'Address Geocoding Cache',
            },
        ),
    ]
